/* Copyright 2011 Jetro Lauha - http://iki.fi/jetro/
 *
 * $Id: MainState.h 373 2011-12-18 23:02:21Z tonic $
 * $Revision: 373 $
 */

#ifndef MAINSTATE_H_INCL
#define MAINSTATE_H_INCL


#include <vector>

//using namespace std;
using namespace irr;
using namespace scene;
using namespace core;
using namespace video;


enum PLAYER_STATE
{
    PS_IDLE = 0, // player did nothing for a while; idle pose
    PS_STILL, // 
    PS_MOVING, // moving
    PS_COUNT //
};

enum PLAYER_SPRITE_DIR
{
    PD_DOWN = 0,
    PD_UP,
    PD_LEFT,
    PD_RIGHT
};


struct Player
{
    Player() : state(PS_STILL), x(5.5f), y(5.0f), prevX(5.5f), prevY(5.0f), idleCount(0), spriteDir(PD_DOWN) {}
    int state;
    float x, y, prevX, prevY;
    int idleCount;
    int spriteDir;
};


class MainState : public AppState
{
public:

    MainState();
    virtual ~MainState();

    // From AppState:

    virtual bool init();
    virtual void deinit();

    virtual bool update(I32 deltaTime);
    virtual void render(irr::IrrlichtDevice *device);

    virtual void onActivate(UI32 time);
    virtual void onDeactivate() {}

    virtual bool onKeyDown(const irr::SEvent::SKeyInput &event);
    virtual bool onKeyUp(const irr::SEvent::SKeyInput &event);
    virtual bool onMouseDown(const irr::SEvent::SMouseInput &event);
    virtual bool onMouseUp(const irr::SEvent::SMouseInput &event);
    virtual bool onMouseMotion(const irr::SEvent::SMouseInput &event);
    virtual bool onMouseWheel(const irr::SEvent::SMouseInput &event);

    void addScore(int amount = 1) { mScore += amount; }

    void gameFinished();

protected:

    void startGame();

    void renderGame(irr::IrrlichtDevice *device);

    void drawSprite(float x, float y, int num, bool horizFlip = false);
    void drawPlayerSprite(float x, float y, STAGE_PLR_TYPE type, int num, bool horizFlip = false);

    enum IMAGE_ID
    {
        IMG_BUTTON = 0,
        IMG_BUTTON_HOT,
        IMG_BUTTON_ACTIVE,
        IMG_BUTTON_HOT_ACTIVE,
        IMG_SPRITES,
        IMG_HEART
    };

    App *mApp;

    IMGUI mIMGUI;
    IMGUIImageButtonStyle *mButtonStyle;

    Font *mFont;
    std::vector<Image *> mImages;

    IrrlichtDevice *mDevice;
    IVideoDriver *mVideo;

    vector2di mLastMousePos;

    bool mTitleUIActive;
    bool mInGame;
    bool mEndingUIActive;

    Player mPlayer;

    Stages mStages;
    int mScore;

    bool mStageChangeScheduled;
    float mFadePhase; // -1..0 = fading state in, 0..1 = fading state out
};


#endif // !MAINSTATE_H_INCL
