/* Copyright 2011 Jetro Lauha - http://iki.fi/jetro/
 *
 * $Id: Stages.cpp 369 2011-12-18 22:46:37Z tonic $
 * $Revision: 369 $
 */

#include <vector>
#include <assert.h>
#include "main.h"
#include "stb_image.c"

using namespace std;

#define ARRAY_COUNT(s) (sizeof(s) / sizeof(s[0]))


static MainState *sMainState;


struct StageGfxLayer
{
    const int y1, x1, y2, x2; // in this order so it's easier to copypaste data from psdparse output
    const char *layerName;
    bool visible;
    int dx, dy; // offset
    bool normalImg; // if true, loaded to image for drawing; if false, loaded as data to pixelData (and visible should be false)
    Color color;
    Image *image;
    stbi_uc *pixelData; // when image==0, then pixelData contains the layer
    int w, h; // only valid when loaded as pixelData!
};


struct Stage
{
    int type; // one of STAGE_TYPE, should match with array index to sStages
    int ctrlMask;
    const char *gfxFileFolder;
    int gfxLayerCount;
    StageGfxLayer *gfxLayers;
    const char *ctrlHelp;
    const char *stagePhaseName, *stageTitle;
    STAGE_PLR_TYPE playerType;
    int plrSpawnX, plrSpawnY;
    bool (*update)(int currentTime, Stage &stage, bool *keyState, bool *metaDataCollState);
    vector<int> data;
};


static StageGfxLayer sGfxLayers_stage_baby[] =
{
    { 192, 280, 282, 414, "room", true, 0, 0, true },
    { 214, 386, 229, 405, "baby-cryhead", false, 0, 0, true },
    { 186, 301, 212, 337, "mother-speechbubble", false, 0, 0, true },
    { 220, 315, 233, 319, "monitorglow", true, 0, 0, true },
    { 205, 326, 214, 335, "heart", false, 0, 0, true },
};

static StageGfxLayer sGfxLayers_stage_school[] =
{
    {   0,   0, 450, 800, "bkg", true, 0, 0, true },
    { 243, 563, 264, 574, "kid3", true, 0, 0, true },
    { 242, 601, 264, 617, "kid4", true, 0, 0, true },
    { 194, 562, 215, 573, "kid1", true, 0, 0, true },
    { 194, 602, 216, 615, "kid2", true, 0, 0, true },
    { 217, 199, 238, 210, "kid-girl", true, 0, 0, true },
    { 220, 201, 223, 208, "girl-joy", false, 0, 0, true },
    { 222, 201, 225, 208, "girl-weep", false, 0, 0, true },
    { 209, 200, 218, 209, "heart", false, 0, 0, true },
    {   0,   0, 450, 800, "METADATA", false, 0, 0, false },
};

static StageGfxLayer sGfxLayers_stage_street[] =
{
    {   0,   0, 450, 800, "Background", true, 0, 0, true },
    { 281,-782, 306, 783, "cars1", true, 0, 0, true },
    { 241,-790, 263, 778, "cars2", true, 0, 0, true },
    { 174,-784, 209, 791, "people1", true, 0, 0, true },
    { 160,  10, 196,1599, "people2", true, 0, 0, true },
    { 297, 597, 336, 636, "lamp1", true, 0, 0, true },
    { 297, 199, 336, 238, "lamp2", true, 0, 0, true },
    { 179, 597, 218, 636, "lamp3", true, 0, 0, true },
    { 179, 199, 218, 238, "lamp4", true, 0, 0, true },
    {   0,   0, 450, 800, "METADATA", false, 0, 0, false },
};

static StageGfxLayer sGfxLayers_stage_homeold[] =
{
    { 182, 256, 272, 466, "room", true, 0, 0, true },
    { 201, 295, 217, 325, "sofa", true, 0, 0, true },
    { 196, 308, 212, 322, "wife", false, 0, 0, true },
    { 244, 297, 247, 323, "tvglow", true, 0, 0, true },
    { 212, 434, 245, 459, "bed", true, 0, 0, true },
    { 214, 437, 238, 457, "sleep", false, 0, 0, true },
    { 171, 289, 197, 339, "speechbubble", false, 0, 0, true },
    {   0,   0, 450, 800, "METADATA", false, 0, 0, false },
};

static bool sUpdate_stage_baby(int currentTime, Stage &stage, bool *keyState, bool *metaDataCollState);
static bool sUpdate_stage_school(int currentTime, Stage &stage, bool *keyState, bool *metaDataCollState);
static bool sUpdate_stage_street(int currentTime, Stage &stage, bool *keyState, bool *metaDataCollState);
static bool sUpdate_stage_homeold(int currentTime, Stage &stage, bool *keyState, bool *metaDataCollState);

static Stage sStages[] =
{
    {
        STAGE_BABY,
        STAGE_CTRL_ACTION,
        "data/stage-baby_png", ARRAY_COUNT(sGfxLayers_stage_baby), sGfxLayers_stage_baby,
        "Z - Cry",
        "Infant", "mom is tired",
        STAGE_PLR_NONE, 400, 225,
        sUpdate_stage_baby,
    },
    {
        STAGE_SCHOOL,
        STAGE_CTRL_MOVE,
        "data/stage-school_png", ARRAY_COUNT(sGfxLayers_stage_school), sGfxLayers_stage_school,
        "Arrows - Move",
        "Childhood", "recess",
        STAGE_PLR_CHILD, 400, 225,
        sUpdate_stage_school,
    },
    {
        STAGE_STREET,
        STAGE_CTRL_MOVE,
        "data/stage-street_png", ARRAY_COUNT(sGfxLayers_stage_street), sGfxLayers_stage_street,
        "Arrows - Move",
        "Adulthood", "grind",
        STAGE_PLR_ADULT, 770, 350,
        sUpdate_stage_street,
    },
    {
        STAGE_HOMEOLD,
        STAGE_CTRL_ALL,
        "data/stage-homeold_png", ARRAY_COUNT(sGfxLayers_stage_homeold), sGfxLayers_stage_homeold,
        "Arrows - Move", // custom-hacked in getStageCtrlHelpString
        "Old age", "still",
        STAGE_PLR_ELDERLY, 270, 235,
        sUpdate_stage_homeold,
    }
};

#define STAGE_COUNT ARRAY_COUNT(sStages)


static float sClamp(float value, float min, float max)
{
    return value < min ? min : (value > max ? max : value);
}

static float sStep(float value, float stepPosition)
{
    return (float)(value < stepPosition ? 0 : 1);
}

static float sBoxStep(float value, float slopeStart, float slopeEnd)
{
    float diff = slopeEnd - slopeStart;
    if (diff == 0)
        return sStep(value, slopeStart);

    return sClamp((value - slopeStart) / diff, 0, 1);
}

static float sBoxPulse(float value, float upSlopeStart, float upSlopeEnd,
                       float downSlopeStart, float downSlopeEnd)
{
    return sBoxStep(value, upSlopeStart, upSlopeEnd) -
           sBoxStep(value, downSlopeStart, downSlopeEnd);
}

static StageGfxLayer * sFindGfxLayer(Stage &stage, const char *name)
{
    assert(name);
    for (int a = 0; a < stage.gfxLayerCount; ++a)
    {
        assert(stage.gfxLayers[a].layerName);
        if (strcmp(stage.gfxLayers[a].layerName, name) == 0)
            return &stage.gfxLayers[a];
    }
    return 0;
}


Stages::Stages()
{
    mCurrentStage = STAGE_BABY;
}


bool Stages::getStageCollisionMetaData(int &w, int &h, unsigned int *&data)
{
    Stage &stage = sStages[mCurrentStage];
    StageGfxLayer *metaData = sFindGfxLayer(stage, "METADATA");
    if (metaData)
    {
        w = metaData->w;
        h = metaData->h;
        // warning: this stuff makes the code endian dependent... see also sStageMetaDataCollValues def.
        data = (unsigned int *)metaData->pixelData;
        return true;
    }
    return false;
}

void Stages::setMainState(class MainState *ms)
{
    // hackish, but the static one is needed for the static update methods...
    mMainState = ms;
    sMainState = ms;
}

bool Stages::loadStageGfxLayers(Graphics *g)
{
    const char *stageImageFormat = ".png";
    for (int a = 0; a < getStageCount(); ++a)
    {
        Stage &stage = sStages[a];
        assert(stage.type == a);
        char imgPath[1000];
        for (int b = 0; b < stage.gfxLayerCount; ++b)
        {
            StageGfxLayer &sgl = stage.gfxLayers[b];
            sprintf(imgPath, "%s/%s%s", stage.gfxFileFolder, sgl.layerName, stageImageFormat);
            if (sgl.normalImg)
            {
                sgl.pixelData = 0;
                sgl.image = g->loadImage(imgPath);
                if (sgl.image == 0)
                {
                    gErrorMessage("Image Load Error", "Can't load %s", imgPath);
                    return false;
                }
            }
            else
            {
                sgl.image = 0;
                int c = 0;
                sgl.pixelData = stbi_load(imgPath, &sgl.w, &sgl.h, &c, 4);
                assert(c == 4);
                assert(sgl.x2 - sgl.x1 == sgl.w);
                assert(sgl.y2 - sgl.y1 == sgl.h);
                if (sgl.pixelData == 0)
                {
                    gErrorMessage("Data Load Error", "Can't load %s", imgPath);
                    return false;
                }
            }
        }
    }
    return true;
}

void Stages::freeStageGfxLayers()
{
    for (int a = 0; a < getStageCount(); ++a)
    {
        Stage &stage = sStages[a];
        for (int b = 0; b < stage.gfxLayerCount; ++b)
        {
            StageGfxLayer &sgl = stage.gfxLayers[b];
            if (sgl.normalImg)
                delete sgl.image;
            else
                stbi_image_free(sgl.pixelData);
        }
    }
}


int Stages::getStageCount()
{
    return STAGE_COUNT;
}

int Stages::getStageCtrlMask()
{
    return sStages[mCurrentStage].ctrlMask;
}

void Stages::setStage(int stage)
{
    assert(stage >= 0 && stage < getStageCount());
    mCurrentStage = stage;
}

void Stages::getStageSpawn(int &x, int &y)
{
    x = sStages[mCurrentStage].plrSpawnX;
    y = sStages[mCurrentStage].plrSpawnY;
}

const char * Stages::getStageCtrlHelpString()
{
    const char *res = sStages[mCurrentStage].ctrlHelp;
    if (mCurrentStage == STAGE_HOMEOLD && sStages[mCurrentStage].data.size() > 1)
    {
        int &nearBed = sStages[mCurrentStage].data[1];
        if (nearBed)
            res = "Z - Sleep";
    }
    return res;
}

const char * Stages::getStagePhaseName()
{
    return sStages[mCurrentStage].stagePhaseName;
}

const char * Stages::getStageTitle()
{
    return sStages[mCurrentStage].stageTitle;
}

STAGE_PLR_TYPE Stages::getStagePlayerType()
{
    STAGE_PLR_TYPE plrType = sStages[mCurrentStage].playerType;
    if (mCurrentStage == STAGE_HOMEOLD && sStages[mCurrentStage].data.size() > 2)
    {
        int &stageDoneFlag = sStages[mCurrentStage].data[2];
        if (stageDoneFlag)
            plrType = STAGE_PLR_NONE; // stop showing player sprite when sleeping in bed
    }
    return plrType;
}


bool Stages::update(int currentTime, bool *keyState, bool *metaDataCollState)
{
    Stage &stage = sStages[mCurrentStage];
    if (stage.update)
        return stage.update(currentTime, stage, keyState, metaDataCollState);
    return false;
}


void Stages::render()
{
    Stage &stage = sStages[mCurrentStage];
    for (int a = 0; a < stage.gfxLayerCount; ++a)
    {
        StageGfxLayer &sgl = stage.gfxLayers[a];
        if (!sgl.visible)
            continue;

        sgl.image->draw(IDM_NORMAL, IF_ALIGN_TOP_LEFT | IF_COLOR | IF_ALPHA, (float)(sgl.x1 + sgl.dx), (float)(sgl.y1 + sgl.dy), sgl.color);
    }
}


static bool sUpdate_stage_baby(int currentTime, Stage &stage, bool *keyState, bool * /*metaDataCollState*/)
{
    assert(keyState);
    if (stage.data.size() == 0)
    {
        // 1st time init
        stage.data.push_back(-1); // cry trigger time
        stage.data.push_back(currentTime); // start
        stage.data.push_back(-1); // heart trigger time
    }

    int &cryTriggerTime = stage.data[0];
    const int stageStartTime = stage.data[1];
    int &heartTriggerTime = stage.data[2];

    const int stage_total_length = 20000;
    const int cry_trigger_length = 5000;
    const int cry_speech_start = 1000;

    StageGfxLayer *monitorGlow = sFindGfxLayer(stage, "monitorglow");
    assert(monitorGlow);
    monitorGlow->color.a = (UI8)(70 + (int)(sin(currentTime * 0.01) * 70) + rand() % 100);

    if ((keyState[K_ACTION] || keyState[K_FLAG_ACTION]) &&
        (cryTriggerTime < 0 || currentTime > cryTriggerTime + cry_trigger_length))
        cryTriggerTime = currentTime;
    StageGfxLayer *cryHead = sFindGfxLayer(stage, "baby-cryhead");
    assert(cryHead);
    StageGfxLayer *speechBubble = sFindGfxLayer(stage, "mother-speechbubble");
    assert(speechBubble);
    speechBubble->visible = false;
    cryHead->visible = false;
    if (keyState[K_ACTION] || (cryTriggerTime >= 0 && currentTime < cryTriggerTime + cry_speech_start))
        cryHead->visible = true;
    if (cryTriggerTime >= 0 &&
        (currentTime > cryTriggerTime + cry_speech_start && currentTime < cryTriggerTime + cry_trigger_length))
    {
        speechBubble->visible = true;
        speechBubble->dy = -(currentTime - cryTriggerTime - cry_speech_start) / 400;
    }

    if (heartTriggerTime >= 0)
    {
        StageGfxLayer *heart = sFindGfxLayer(stage, "heart");
        assert(heart);
        heart->visible = true;
        heart->dy = -(currentTime - heartTriggerTime) / 100;
    }

    if (currentTime >= stageStartTime + stage_total_length - 500 &&
        cryTriggerTime < 0 &&
        heartTriggerTime < 0)
    {
        heartTriggerTime = currentTime; // get heart by staying quiet
        sMainState->addScore();
    }

    if (currentTime >= stageStartTime + stage_total_length)
    {
        // stage done
        return true;
    }

    return false;
}

static bool sUpdate_stage_school(int currentTime, Stage &stage, bool *keyState, bool *metaDataCollState)
{
    assert(keyState);
    if (stage.data.size() == 0)
    {
        // 1st time init
        stage.data.push_back(currentTime); // start
        stage.data.push_back(-1); // heart trigger time
        stage.data.push_back(0); // girl cry triggered (0/1)
        stage.data.push_back(-1); // girl friendship meter start time
        stage.data.push_back(-1); // last frame time for stage
        stage.data.push_back(-1); // kid1 movement time
        stage.data.push_back(0); // kid1 movement visit-x
        stage.data.push_back(0); // kid1 movement visit-y
        stage.data.push_back(-1); // kid2 movement time
        stage.data.push_back(0); // kid2 movement visit-x
        stage.data.push_back(0); // kid2 movement visit-y
        stage.data.push_back(-1); // kid3 movement time
        stage.data.push_back(0); // kid3 movement visit-x
        stage.data.push_back(0); // kid3 movement visit-y
        stage.data.push_back(-1); // kid4 movement time
        stage.data.push_back(0); // kid4 movement visit-x
        stage.data.push_back(0); // kid4 movement visit-y
    }

    const int stageStartTime = stage.data[0];
    int &lastTime = stage.data[4];
    int &heartTriggerTime = stage.data[1];
    int &girlCryTriggered = stage.data[2];
    int &girlFriendshipTimer = stage.data[3];
    const int deltaTime = currentTime - lastTime;

    const int otherKidCount = 4;
    int idx = 5;
    int *kidtime[otherKidCount] = { &stage.data[idx], &stage.data[idx + 3], &stage.data[idx + 6], &stage.data[idx + 9] };
    ++idx;
    int *kidvx[otherKidCount] = { &stage.data[idx], &stage.data[idx + 3], &stage.data[idx + 6], &stage.data[idx + 9] };
    ++idx;
    int *kidvy[otherKidCount] = { &stage.data[idx], &stage.data[idx + 3], &stage.data[idx + 6], &stage.data[idx + 9] };

    const int stage_total_length = 15000;

    StageGfxLayer *kid[4] = { sFindGfxLayer(stage, "kid1"), sFindGfxLayer(stage, "kid2"), sFindGfxLayer(stage, "kid3"), sFindGfxLayer(stage, "kid4") };
    assert(kid[0] && kid[1] && kid[2] && kid[3]);

    StageGfxLayer *kidgirl = sFindGfxLayer(stage, "kid-girl");
    StageGfxLayer *girljoy = sFindGfxLayer(stage, "girl-joy");
    StageGfxLayer *girlweep = sFindGfxLayer(stage, "girl-weep");
    assert(kidgirl && girljoy && girlweep);

    const int timeToMakeFriends = 4000;
    const int timeForGirlJoy = 750;
    if (metaDataCollState[SMCI_TRIG_POS])
    {
        if (girlFriendshipTimer < 0)
            girlFriendshipTimer = currentTime;
        if (girlFriendshipTimer >= 0 &&
            currentTime > girlFriendshipTimer + timeToMakeFriends &&
            !girlCryTriggered && // can't get heart if she weeps
            heartTriggerTime < 0)
        {
            heartTriggerTime = currentTime;
            sMainState->addScore(); // get heart by trying to make friends
            // enable wife in homeold-stage
            Stage &stage_homeOld = sStages[STAGE_HOMEOLD];
            StageGfxLayer *wife = sFindGfxLayer(stage_homeOld, "wife");
            if (wife)
                wife->visible = true;
        }
        if (girlFriendshipTimer >= 0 &&
            currentTime > girlFriendshipTimer + timeForGirlJoy)
        {
            girljoy->visible = true;
            girlweep->visible = false; // joy still cancels weep even if no heart
        }
    }
    else
        girlFriendshipTimer = -1;

    if (metaDataCollState[SMCI_TRIG_NEG])
    {
        girlCryTriggered = 1;
        girljoy->visible = false;
        if (heartTriggerTime < 0) // no need to make her cry if you made friends
            girlweep->visible = true;
    }

    for (int a = 0; a < otherKidCount; ++a)
    {
        assert(kidtime[a] && kidvx[a] && kidvy[a] && kid[a]);
        int thisKidMaxTime = 1000 + a * 100; // a bit different movement phase for the other kids
        if (*kidtime[a] > thisKidMaxTime || *kidtime[a] < 0)
        {
            *kidtime[a] = 0;
            *kidvx[a] = rand() % 21 - 10;
            *kidvy[a] = rand() % 21 - 10;
        }
        float vPhase = sBoxPulse((float)*kidtime[a], 0, thisKidMaxTime / 2.f, thisKidMaxTime / 2.f, (float)thisKidMaxTime);
        kid[a]->dx = (int)(*kidvx[a] * vPhase);
        kid[a]->dy = (int)(*kidvy[a] * vPhase);
        // only move other kids when you're not trying to interfere
        if (!metaDataCollState[SMCI_TRIG_NEG])
            *kidtime[a] += deltaTime;
    }

    if (heartTriggerTime >= 0)
    {
        StageGfxLayer *heart = sFindGfxLayer(stage, "heart");
        assert(heart);
        if (currentTime < heartTriggerTime + 4000)
            heart->visible = true;
        else
            heart->visible = false;
        heart->dy = -(currentTime - heartTriggerTime) / 100;
    }

    lastTime = currentTime;
    if (currentTime >= stageStartTime + stage_total_length)
    {
        // stage done
        return true;
    }

    return false;
}

static bool sUpdate_stage_street(int currentTime, Stage &stage, bool *keyState, bool *metaDataCollState)
{
    if (stage.data.size() == 0)
    {
        // 1st time init
        stage.data.push_back(currentTime); // start
        stage.data.push_back(0); // stage done flag (0/1)
    }

    const int stageStartTime = stage.data[0];
    int &doneFlag = stage.data[1];

    StageGfxLayer *cars1 = sFindGfxLayer(stage, "cars1");
    StageGfxLayer *cars2 = sFindGfxLayer(stage, "cars2");
    StageGfxLayer *people1 = sFindGfxLayer(stage, "people1");
    StageGfxLayer *people2 = sFindGfxLayer(stage, "people2");

    cars1->dx = (currentTime / 12) % WIDTH;
    cars2->dx = (currentTime / 7) % WIDTH;
    people1->dx = (currentTime / 41) % WIDTH;
    people2->dx = -(currentTime / 40) % WIDTH;

    if (metaDataCollState[SMCI_TRIG_1])
    {
        doneFlag = 1;
        return true; // stage done
    }

    return false;
}

static bool sUpdate_stage_homeold(int currentTime, Stage &stage, bool *keyState, bool *metaDataCollState)
{
    /*
    static StageGfxLayer sGfxLayers_stage_homeold[] =
{
    { 182, 256, 272, 466, "room", true, 0, 0, true },
    { 201, 295, 217, 325, "sofa", true, 0, 0, true },
    { 196, 308, 212, 322, "wife", false, 0, 0, true },
    { 244, 297, 247, 323, "tvglow", true, 0, 0, true },
    { 212, 434, 245, 459, "bed", true, 0, 0, true },
    { 214, 437, 268, 578, "sleep", false, 0, 0, true },
    { 171, 289, 197, 339, "speechbubble", false, 0, 0, true },
    {   0,   0, 450, 800, "METADATA", false, 0, 0, false },
};
*/
    assert(keyState);
    if (stage.data.size() == 0)
    {
        // 1st time init
        stage.data.push_back(currentTime); // start
        stage.data.push_back(0); // near bed flag 0/1
        stage.data.push_back(0); // stage done (sleep) flag 0/1
    }

    const int stageStartTime = stage.data[0];
    int &nearBed = stage.data[1];
    int &stageDoneFlag = stage.data[2];
    const int prevStageDoneFlag = stageDoneFlag;

    StageGfxLayer *tvglow = sFindGfxLayer(stage, "tvglow");
    assert(tvglow);
    tvglow->color.a = (UI8)(70 + (int)(sin(currentTime * 0.01) * 70) + rand() % 100);

    StageGfxLayer *wife = sFindGfxLayer(stage, "wife");
    StageGfxLayer *sleep = sFindGfxLayer(stage, "sleep");
    StageGfxLayer *speechbubble = sFindGfxLayer(stage, "speechbubble");
    assert(wife && sleep && speechbubble);

    if (metaDataCollState[SMCI_TRIG_1])
        nearBed = 1;
    else
        nearBed = 0;

    if ((keyState[K_ACTION] || keyState[K_FLAG_ACTION]) && nearBed)
    {
        stageDoneFlag = 1;
        sleep->visible = true;
    }

    speechbubble->visible = false;
    if (metaDataCollState[SMCI_TRIG_NEG] && wife->visible && !stageDoneFlag)
    {
        speechbubble->visible = true;
    }


    if (prevStageDoneFlag == 0 && stageDoneFlag == 1)
    {
        // just finished
        sMainState->gameFinished();
    }

    // we actually want end UI just keep showing this state
    return false; // stageDoneFlag == 1;
}
