/* Copyright 2011 Jetro Lauha - http://iki.fi/jetro/
 *
 * $Id: Stages.h 369 2011-12-18 22:46:37Z tonic $
 * $Revision: 369 $
 */

#ifndef STAGES_H
#define STAGES_H


enum STAGE_CTRLMASK
{
    STAGE_CTRL_ACTION = 0x01,
    STAGE_CTRL_MOVE   = 0x02,
    STAGE_CTRL_ALL = STAGE_CTRL_ACTION | STAGE_CTRL_MOVE
};

enum STAGE_TYPE
{
    STAGE_BABY = 0,
    STAGE_SCHOOL,
    STAGE_STREET,
    STAGE_HOMEOLD
};

enum STAGE_PLR_TYPE
{
    STAGE_PLR_NONE = -1,
    STAGE_PLR_CHILD = 0,
    STAGE_PLR_ADULT,
    STAGE_PLR_ELDERLY
};

enum STAGE_METADATA_COLL_MASK
{
    SMC_MASK_ALPHA = 0xff000000,
    SMC_MASK_COLOR = 0x00ffffff
};

enum STAGE_METADATA_COLL_IDX
{
    SMCI_WALL = 0,
    SMCI_TRIG_POS,
    SMCI_TRIG_NEG,
    SMCI_TRIG_1,
    SMCI_COUNT //
};

static const int sStageMetaDataCollValues[] =
{
    // stbi_load returns bytes in r,g,b,a order, which we cast to uint32, so values below are 0xAABBGGRR
    0x000000ff, // SMC_WALL
    0x0000ff00, // SMC_TRIG_POS
    0x00ff0000, // SMC_TRIG_NEG
    0x00ff00ff, // SMC_TRIG_1
};


class Stages
{
public:

    Stages();

    void setMainState(class MainState *ms);

    bool loadStageGfxLayers(Graphics *g);
    void freeStageGfxLayers();
    void setStage(int stageType);

    int getStageCount();
    int getStageCtrlMask();

    int getCurrentStage() { return mCurrentStage; }

    void getStageSpawn(int &x, int &y);
    const char * getStageCtrlHelpString();
    const char * getStagePhaseName();
    const char * getStageTitle();
    STAGE_PLR_TYPE getStagePlayerType();

    // returns true if there is collision metadata, false if not
    bool getStageCollisionMetaData(int &w, int &h, unsigned int *&data);

    // should return true when stage is done and we should advance to next stage
    bool update(int currentTime, bool *keyState, bool *metaDataCollState);
    void render();

    int mCurrentStage;
    class MainState *mMainState;
};


#endif // STAGES_H
